from dataclasses import dataclass
from typing import Dict, Optional, Tuple
from pararun.protocol.queue_client_protocol import QueueProtocol
from pararun.transport.serializers import JsonSerializer


@dataclass
class Adapter:
    adapter_protocol: QueueProtocol
    override_function: Optional[Tuple[str,str]] = None
    override_batcher: Optional[Tuple[Optional[str],Optional[str], int, int, int]] = None
    init_function: Optional[Tuple[str, str]] = None
    name: Optional[str] = None

    def assure_serialization_compatibility(self, x):
        serializer = JsonSerializer()
        x = serializer.serialize(x)
        y = serializer.deserialize(x)
        return y


def build(origin, object=None, event_id=None, profile_id=None, flow_id=None, node_id=None, user_id=None, error_number=None, traceback=None) -> dict:
    return dict(
        origin=origin,
        class_name=object.__class__.__name__ if object else None,
        package=object.__class__.__module__ if object else None,
        event_id=event_id,
        profile_id=profile_id,
        flow_id=flow_id,
        node_id=node_id,
        user_id=user_id,
        error_number=error_number
    )

def exact(origin, event_id=None, profile_id=None, flow_id=None, node_id=None, class_name=None, package=None, user_id=None, error_number=None, traceback=None) -> dict:
    return dict(
        origin=origin,
        class_name=class_name,
        package=package,
        event_id=event_id,
        profile_id=profile_id,
        flow_id=flow_id,
        node_id=node_id,
        user_id=user_id,
        error_number=error_number
    )



import inspect
from contextlib import contextmanager

from pararun.model.transport_context import TransportContext
from pararun.service.converters.time_converters import pretty_time_format
from time import time
from importlib import import_module

from pararun.service.logger.extra_info import exact
from pararun.service.logger.log_handler import get_logger

logger = get_logger(__name__)
_cached_function_references = {}


@contextmanager
def single_function_reference(module, name):
    key = (module, name)
    if key not in _cached_function_references:
        module_ref = import_module(module)
        func_ref = getattr(module_ref, name)
        _cached_function_references[key] = func_ref

    yield _cached_function_references[key]


def invoke(context: TransportContext, module, name, args, kwargs=None):
    with single_function_reference(module, name) as func_ref:
        if kwargs is None:
            kwargs = {}
        # This is the signature of a function
        return func_ref(context, *args, **kwargs)


async def async_invoke(context: TransportContext, module, name, args, kwargs=None):
    t = time()

    result = invoke(context, module, name, args, kwargs)
    if inspect.iscoroutine(result):
        result = await result

    logger.debug(f"Invoked {module}.{name}. Finished in {pretty_time_format(time() - t)}",
                 extra=exact(origin="background-worker", package=__name__, class_name='invoke'))

    return result


async def raw_func_invoke(module, name, args=None, kwargs=None):
        t = time()

        with single_function_reference(module, name) as func_ref:
            if kwargs is None:
                kwargs = {}
            # This is the signature of a function
            result = func_ref(*args, **kwargs)

        if inspect.iscoroutine(result):
            result = await result

        logger.debug(f"Invoked {module}.{name}. Finished in {pretty_time_format(time() - t)}",
                     extra=exact(origin="background-worker", package=__name__, class_name='invoke'))

        return result


import os

from pararun.service.environment import get_env_as_bool
from pararun.service.singleton import Singleton


class Config(metaclass=Singleton):

    @property
    def fail_over_server_api(self):
        return os.environ.get('FAIL_OVER_SERVER_API', None)

    @property
    def fail_over_server_token(self):
        return os.environ.get('FAIL_OVER_SERVER_TOKEN', None)

    @property
    def queue_enabled(self):
        return get_env_as_bool('QUEUE_ENABLED', 'yes')


from typing import Protocol


class MessageProtocol(Protocol):

    def value(self):
        pass

    def properties(self) -> dict:
        pass

    def message(self):
        pass

    def raw(self):
        pass

    def offset(self):
        pass

    def partition(self):
        pass

    def topic(self):
        pass


import os


def get_env_as_int(env_key, default_value):
    value = os.environ.get(env_key, default_value)
    if not value:
        return default_value
    try:
        return int(value)
    except Exception:
        return default_value


def _str_to_bool(value):
    if isinstance(value, bool):
        return value

    true_values = ['yes', 'on', 'true']
    false_values = ['no', 'off', 'false']
    value_lower = value.lower()
    if value_lower in true_values:
        return True
    elif value_lower in false_values:
        return False
    else:
        raise ValueError(f"Invalid value for boolean setting: {value}")


def get_env_as_bool(env_key, default_value) -> bool:
    value = os.environ.get(env_key, default_value)
    return _str_to_bool(value)



import json
import logging

from pararun.service.logger.tools import Q_INFO_LEVEL, DEV_INFO_LEVEL, Q_STAT_LEVEL


class CustomFormatter(logging.Formatter):
    white = "\x1b[97m"
    grey = "\x1b[37m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    blue = "\x1b[34m"
    magenta = "\x1b[35m"
    lavender = "\x1b[38;5;183m"

    _format_str = "%(asctime)s [%(levelname)s] [%(error_number)s] %(message)s | %(name)s | %(filename)s | %(lineno)d"

    FORMATS = {
        logging.DEBUG: grey + _format_str + reset,
        DEV_INFO_LEVEL: magenta + _format_str + reset,
        Q_INFO_LEVEL: blue + _format_str + reset,
        Q_STAT_LEVEL: lavender + _format_str + reset,
        logging.INFO: white + _format_str + reset,
        logging.WARNING: yellow + _format_str + reset,
        logging.ERROR: red + _format_str + reset,
        logging.CRITICAL: bold_red + _format_str + reset
    }

    def format(self, record):
        if not hasattr(record, "error_number") or record.error_number is None:
            record.error_number = "N/A"
        # fallback to _format_str instead of self.format
        log_fmt = self.FORMATS.get(record.levelno, self._format_str)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


class ConsoleFormatter(logging.Formatter):
    _format_str = "%(asctime)s [%(levelname)s] [%(error_number)s] %(message)s | %(name)s | %(filename)s | %(lineno)d"

    FORMATS = {
        logging.DEBUG: _format_str,
        DEV_INFO_LEVEL: _format_str,
        Q_INFO_LEVEL: _format_str,
        Q_STAT_LEVEL: _format_str,
        logging.INFO: _format_str,
        logging.WARNING: _format_str,
        logging.ERROR: _format_str,
        logging.CRITICAL: _format_str
    }

    def format(self, record):
        if not hasattr(record, "error_number") or record.error_number is None:
            record.error_number = "N/A"
        log_fmt = self.FORMATS.get(record.levelno, self._format_str)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


class JSONFormatter(logging.Formatter):
    def format(self, record):
        # Ensure asctime and message are computed
        record.asctime = self.formatTime(record, self.datefmt)
        record.message = record.getMessage()

        log_record = {
            "timestamp": record.asctime,
            "level": record.levelname,
            "message": record.message,
            "name": record.name,
            "filename": record.filename,
            "lineno": record.lineno,
        }
        return json.dumps(log_record)



from typing import Optional, Callable

import functools
import contextlib

from pararun.model.worker_capsule import WorkerCapsule, FunctionCapsule
from pararun.service.logger.log_handler import get_logger

logger = get_logger(__name__)


def deferred_function(module: Optional[str] = None,
                      guard: Optional[Callable] = None,
                      guard_module: Optional[str] = None):
    def decorator_serialize_function(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):

            nonlocal module, guard, guard_module

            # Instead of executing the function, return a WorkerCapsule object
            if module is None:
                module = func.__module__

            capsule = WorkerCapsule(
                function=FunctionCapsule(name=func.__name__, module=module),
                args=args,
                kwargs=kwargs
            )

            if guard:
                if guard_module is None:
                    guard_module = guard.__module__
                capsule.guard = FunctionCapsule(name=guard.__name__, module=guard_module)

            return capsule

        return wrapper

    return decorator_serialize_function


def deferred_lazy_function(module: str,
                           func: str,
                           guard: Optional[Callable] = None,
                           guard_module: Optional[str] = None):
    def wrapper(*args, **kwargs):

        nonlocal module, guard, guard_module

        # Instead of executing the function, return a WorkerCapsule object

        capsule = WorkerCapsule(
            function=FunctionCapsule(name=func, module=module),
            args=args,
            kwargs=kwargs
        )

        if guard:
            if guard_module is None:
                guard_module = guard.__module__
            capsule.guard = FunctionCapsule(name=guard.__name__, module=guard_module)

        return capsule

    return wrapper


@contextlib.contextmanager
def deferred_execution(module: Optional[str] = None,
                       guard: Optional[Callable] = None,
                       guard_module: Optional[str] = None):
    yield deferred_function(
        module,
        guard=guard,
        guard_module=guard_module
    )


@contextlib.contextmanager
def deferred_lazy_execution(module: str,
                            func: str,
                            guard: Optional[Callable] = None,
                            guard_module: Optional[str] = None):
    yield deferred_lazy_function(
        module,
        func,
        guard=guard,
        guard_module=guard_module
    )



import os

from pararun.service.logger.formater import ConsoleFormatter, JSONFormatter, CustomFormatter


def log_format_adapter():
    type = os.environ.get('LOGGING_FORMAT', 'console')
    if type == 'console':
        return ConsoleFormatter()
    elif type == 'json':
        return JSONFormatter()
    else:
        return CustomFormatter()


from datetime import datetime, timedelta

from pararun.service.singleton import Singleton


class TimedFlag:
    def __init__(self):
        self._value = False
        self._expiry_time = None

    def set_false(self):
        self._value = False

    def set_true(self, ttl_seconds):
        """
        Set the value to True and define the time-to-live (TTL) in seconds.
        """
        self._value = True
        self._expiry_time = datetime.now() + timedelta(seconds=ttl_seconds)

    def get_value(self):
        """
        Return the current value of the flag. If the TTL has expired, reset to False.
        """
        if self._value and self._expiry_time <= datetime.now():
            self._value = False
            self._expiry_time = None
        return self._value

    def timeout(self):
        return self._expiry_time - datetime.now()


class TimedFlags(metaclass=Singleton):
    def __init__(self):
        self._data = {}

    def set(self, key, value, ttl):
        """
        Set a timed flag for the given key with the specified TTL.
        """
        if not isinstance(value, bool):
            raise ValueError("Flag can only be bool.")

        # Retrieve existing or create a new TimedFlag
        flag: TimedFlag = self._data.get(key, TimedFlag())
        if value:
            flag.set_true(ttl)
        else:
            flag.set_false()

        # Store back the modified flag
        self._data[key] = flag

    def get(self, key) -> bool:
        """
        Get the value of the timed flag for the given key.
        """
        if key not in self._data:
            self._data[key] = TimedFlag()
        flag = self._data.get(key)
        return flag.get_value()


    def get_flag(self, key) -> TimedFlag:
        """
        Get the value of the timed flag for the given key.
        """
        if key not in self._data:
            return TimedFlag()
        return self._data.get(key)



from pararun.config import Config
import requests
from pararun.service.logger.log_handler import get_logger

config = Config()

logger = get_logger(__name__)

def fallback_on_error(payload, adapter_name: str):
    fail_over_server_api = config.fail_over_server_api
    fail_over_server_token = config.fail_over_server_token

    if fail_over_server_api is None or fail_over_server_token is None:
        logger.info("Fail-over server not configured.")
        return

    response = requests.post(
        fail_over_server_api,
        headers={'Authorization': f'Bearer {fail_over_server_token}'},
        json={
            "headers": {
                'x-adapter': adapter_name,
                'x-restore': 'pulsar'
            },
            "payload": payload.model_dump(mode='json')
        }
    )

    if response.status_code != 200:
        raise ConnectionError(f"Could not connect to fail over server {fail_over_server_api}.")


import logging

DEV_INFO_LEVEL = 13
Q_INFO_LEVEL = 15
Q_STAT_LEVEL = 25

def _get_logging_level(level: str) -> int:
    level = level.upper()
    if level == 'DEBUG':
        return logging.DEBUG
    if level == 'DEV_INFO':
        return DEV_INFO_LEVEL
    if level == 'Q_INFO':
        return Q_INFO_LEVEL
    if level == 'INFO':
        return logging.INFO
    if level == 'STAT':
        return Q_STAT_LEVEL
    if level == 'WARNING' or level == "WARN":
        return logging.WARNING
    if level == 'ERROR':
        return logging.ERROR
    if level == 'CRITICAL':
        return logging.CRITICAL

    return logging.WARNING


from dataclasses import dataclass
from typing import Callable, Optional, Tuple, Union


@dataclass
class BatcherConfig:
    func: Union[Callable, str]
    timeout: float
    module: Optional[str] = None
    min_size: Optional[int] = 0
    max_size: Optional[int] = 0

    def get_module_and_function(self) -> Tuple[str, str]:
        if self.module is None:
            batcher_module = self.func.__module__
        else:
            batcher_module = self.module

        if isinstance(self.func, Callable):
            batcher_func = self.func.__name__
        else:
            batcher_func = self.func


        return batcher_module, batcher_func



from typing import Protocol, Tuple, Any

from pararun.model.transport_context import TransportContext


class SerializerProtocol(Protocol):

    def __init__(self, schema=None):
        pass

    def serialize(self, data, event_name: str, context: TransportContext) -> Tuple[Any, Any]:
        # Takes data serializes it and construct a pulsar Record
        pass

    def deserialize(self, record) -> Tuple[tuple, dict, dict]:
        # Takes object and unserialize the params and returns Function, args, kwargs, context
        pass

    def schema(self):
        pass

    def get_serializer(self):
        pass



from typing import Protocol, Callable

from pararun.model.data_bus import DataBus
from pararun.protocol.queue_consumer_protocol import ConsumerProtocol

QUEUE_OK = 202
QUEUE_BACK_PRESSURE = 429
MESSAGE_TOO_LARGE = 413

class QueueProtocol(Protocol):

    def publish(self, payload, on_error: Callable) ->int:  # Should status
        pass

    def consumer(self) -> ConsumerProtocol:
        pass

    def data_bus(self)->DataBus:
        pass

    def close(self):
        pass

    def flush(self):
        pass



import random

from pararun.service.timed_flag import TimedFlags
from pararun.service.logger.log_handler import get_logger
from pararun.service.singleton import Singleton

logger = get_logger(__name__)

class FallbackManager(metaclass=Singleton):

    KEY = 'pulsar_error'

    def __init__(self):
        self.flags = TimedFlags()

    def set_error_mode(self, message):
        fallback_time = random.uniform(15, 30)
        logger.warning(
            f"Falling to inline invocation due to queue error. Next check if service is back: {fallback_time}s. Details: {message}")
        self.flags.set(FallbackManager.KEY, True, fallback_time)

    def is_in_error_mode(self):
        return self.flags.get(FallbackManager.KEY)



from dataclasses import dataclass, asdict
from typing import Optional, Any, Union
from uuid import uuid4

from pararun.protocol.model_factory_protocol import SerializerProtocol


@dataclass
class DataBusSubscription:
    subscription_name: str
    consumer_name: str
    receiver_queue_size: int
    consumer_type:  Optional[Union[int,str]]
    initial_position: Optional[Union[int,str]]


@dataclass
class DataBus:
    topic: str
    factory: Optional[SerializerProtocol]
    subscription: DataBusSubscription

    def get_subscription_settings_as_dict(self) -> dict:
        _dict = asdict(self.subscription)
        _dict['consumer_name'] = f"{self.subscription.consumer_name}-{str(uuid4())[:6]}"
        if self.factory:
            _dict['schema'] = self.factory.schema()

        return _dict



import json
import os
import sys

import logging
from datetime import datetime
from typing import List
from zoneinfo import ZoneInfo

from logging import Handler, LogRecord
from time import time

from pararun.service.logger.logger_adapter import log_format_adapter
from pararun.service.logger.logging_config import log_stack_trace_as, log_stack_trace_for
from pararun.service.logger.tools import _get_logging_level, Q_INFO_LEVEL, DEV_INFO_LEVEL, Q_STAT_LEVEL
import traceback

_env = os.environ
_logging_level = _get_logging_level(_env['LOGGING_LEVEL']) if 'LOGGING_LEVEL' in _env else logging.WARNING
_save_logs_on = _env.get('SAVE_LOGS', 'yes').lower() == 'yes'
_log_format_adapter = log_format_adapter()


logging.addLevelName(Q_INFO_LEVEL, "Q_INFO")
logging.addLevelName(DEV_INFO_LEVEL, "DEV_INFO")
logging.addLevelName(Q_STAT_LEVEL, "STAT")

def now_in_utc():
    return datetime.utcnow().replace(tzinfo=ZoneInfo('UTC'))


def stack_trace(level):
    if level not in log_stack_trace_for:
        return {}

    # Extract the traceback object
    tb = sys.exc_info()[2]

    # Convert the traceback to a list of structured frames
    stack = traceback.extract_tb(tb)

    if not stack:
        stack = traceback.extract_stack()

    metadata = {}

    # Format the stack trace as a list of dictionaries
    return {
        "context": metadata,
        "stack": [
            {
                "filename": frame.filename,
                "line_number": frame.lineno,
                "function_name": frame.name,
                "code_context": frame.line
            }
            for frame in stack
        ]}


class DeferLogHandler(Handler):

    def __init__(self, level=0, collection=None):
        super().__init__(level)
        if collection is None:
            collection = []
        self.collection: List = collection
        self.last_save = time()

    @staticmethod
    def _get(record, value, default_value):
        return record.__dict__.get(value, default_value)

    def emit(self, record: LogRecord):

        # Skip info and debug.
        if record.levelno <= logging.INFO:
            return

        _trace = stack_trace(record.levelname)
        if log_stack_trace_as == 'json':
            if _trace:
                stack_trace_str = f"JSON:{json.dumps(_trace)}"
            else:
                stack_trace_str = None
        else:
            stack_trace_str = record.stack_info

        log = {  # Maps to tracardi-log index
            "date": now_in_utc(),
            "message": record.msg,
            "logger": record.name,
            "file": record.filename,
            "line": record.lineno,
            "level": record.levelname,
            "stack_info": stack_trace_str,
            # "exc_info": record.exc_info  # Can not save this to TrackerPayload
            "module": self._get(record, "package", record.module),
            "class_name": self._get(record, "class_name", record.funcName),
            "origin": self._get(record, "origin", "root"),
            "event_id": self._get(record, "event_id", None),
            "profile_id": self._get(record, "profile_id", None),
            "flow_id": self._get(record, "flow_id", None),
            "node_id": self._get(record, "node_id", None),
            "user_id": self._get(record, "user_id", None),
        }

        if _save_logs_on:
            self.collection.append(log)

    def has_logs(self, min_log_size=None):
        if min_log_size is None:
            min_log_size = 100
        if not isinstance(self.collection, list):
            return False
        return len(self.collection) >= min_log_size or (time() - self.last_save) > 60

    def reset(self):
        self.collection = []
        self.last_save = time()


class StackInfoLogger(logging.Logger):
    def error(self, msg, *args, **kwargs):
        kwargs['stack_info'] = True
        if 'exc_info' not in kwargs:
            kwargs['exc_info'] = True
        super().error(msg, *args, **kwargs)

    def critical(self, msg, *args, **kwargs):
        kwargs['stack_info'] = True
        kwargs['exc_info'] = True
        super().error(msg, *args, **kwargs)

    def q_info(self, msg, *args, **kwargs):
        if self.isEnabledFor(Q_INFO_LEVEL):
            self._log(Q_INFO_LEVEL, msg, args, **kwargs)

    def stat(self, msg, *args, **kwargs):
        if self.isEnabledFor(Q_STAT_LEVEL):
            self._log(Q_STAT_LEVEL, msg, args, **kwargs)

    def dev_info(self, msg, *args, **kwargs):
        if self.isEnabledFor(DEV_INFO_LEVEL):
            self._log(DEV_INFO_LEVEL, msg, args, **kwargs)


logging.setLoggerClass(StackInfoLogger)
logging.basicConfig(level=logging.INFO)


def get_logger(name, level=None):
    # Replace the default logger class with your custom class
    logger = logging.getLogger(name)
    logger.propagate = False
    logger.setLevel(level or _logging_level)

    # Elastic log formatter

    logger.addHandler(log_handler)

    # Console log handler

    clh = logging.StreamHandler()
    clh.setFormatter(_log_format_adapter)
    logger.addHandler(clh)

    return logger


log_handler = DeferLogHandler()



from typing import Protocol, Generator

from pararun.protocol.queue_message_protocol import MessageProtocol


class ConsumerProtocol(Protocol):

    def receive(self, timeout_millis) -> Generator[MessageProtocol, None, None]:
        """
        Raises ClientTimeOutError if timeout_millis passed
        :param timeout_millis:
        :return:
        """
        pass

    def acknowledge(self, msg: MessageProtocol):
        pass

    # Runs after batch
    def commit(self, msg: MessageProtocol):
        pass

    def negative_acknowledge(self, msg: MessageProtocol):
        pass

    def can_ack(self) -> bool:
        pass


from pararun.model.adapter import Adapter
from pararun.protocol.queue_client_protocol import QueueProtocol
from pararun.service.singleton import Singleton
from pararun.transport.serializers import JsonSerializer

_default_serializer = JsonSerializer

class MockAdapter(QueueProtocol):

    def consumer(self, record):
        return _default_serializer.deserialize(record)

    def publish(self, payload, on_error):
        return _default_serializer.serialize(payload)

    def data_bus(self):
        pass

    def close(self):
        pass

    def flush(self):
        pass

class DeferAdapterSelector(metaclass=Singleton):

    def get(self, adapter_name, queue_tenant: str):
        return Adapter(adapter_protocol=MockAdapter())



import os

from pararun.service.environment import get_env_as_int

log_stack_trace_as = os.environ.get('LOG_STACK_TRACE_AS', 'json')
log_stack_trace_for = os.environ.get('LOG_STACK_TRACE_AS', 'CRITICAL,ERROR,WARNING,INFO').split(',')
log_bulk_size = get_env_as_int('LOG_BULK_SIZE', 1)


from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='airembr_pararun_os',
    version='0.0.8',
    description='Airembr Pararun Module',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Risto Kowaczewski',
    packages=['pararun', 'pararun_adapter'],
    install_requires=[
        'pydantic',
        'requests',
        'orjson',
        'durable-dot-dict>=0.0.24'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    keywords=['airembr', 'sdk'],
    include_package_data=True,
    python_requires=">=3.10",
)



def pretty_time_format(time_in_seconds: float) -> str:
    hours = int(time_in_seconds // 3600)
    minutes = int((time_in_seconds % 3600) // 60)
    seconds = int(time_in_seconds % 60)
    milliseconds = int((time_in_seconds % 1) * 1000)

    if hours > 0:
        return f"{hours:02}:{minutes:02}:{seconds:02}.{milliseconds:03}"
    else:
        return f"{minutes:02}:{seconds:02}.{milliseconds:03}"



FUNCTION = 'system.function'
STORAGE = 'system.storage'
COLLECTOR = 'system.collector'
WORKFLOW = 'system.workflow'
DESTINATION = 'system.destination'
LOGGER = 'system.logger'
PROPERTIES = 'system.event.properties'
ATTACHMENTS = 'system.event.attachments'
ENTITY_MERGING = 'system.entity.merging'


from typing import Optional, Callable, Any
from pydantic import BaseModel

from pararun.config import Config
from pararun.consumer.batcher import BulkedResult
from pararun.model.batcher import BatcherConfig
from pararun.model.status import DispatchStatus
from pararun.model.transport_context import TransportContext
from pararun.service.fallback import FallbackManager
from pararun.service.invokers import invoke, async_invoke
from pararun.service.logger.log_handler import get_logger, log_handler

logger = get_logger(__name__)
fallback = FallbackManager()
config = Config()


class FunctionCapsule(BaseModel):
    module: str
    name: str


class PublishPayload(BaseModel):
    capsule: 'WorkerCapsule'
    job_tag: str
    context: TransportContext
    headers: dict
    options: Optional[dict] = {}


class WorkerCapsule(BaseModel):
    function: FunctionCapsule
    args: tuple
    kwargs: dict
    guard: Optional[FunctionCapsule] = None

    def _allow(self, context: TransportContext):
        if not self.guard:
            return True
        return invoke(context, self.guard.module, self.guard.name, self.args, self.kwargs)

    async def run(self, context: TransportContext):

        if not self._allow(context):
            return None

        result = await async_invoke(context, self.function.module, self.function.name, self.args, self.kwargs)

        return result

    @staticmethod
    def _convert(result):
        for _result in result:
            if isinstance(_result, BaseModel):
                yield _result.model_dump(mode='json')
            else:
                yield _result


    async def _invoke(self, batcher, context):

        # Run as asyncio task without batcher
        result = await self.run(context)

        if isinstance(result, BulkedResult):
            result = result.result

        if not batcher:
            return result

        # With batcher
        if not isinstance(result, list):
            result = [result]

        result = list(self._convert(result))

        batcher_module, batcher_name = batcher.get_module_and_function()
        return await async_invoke(context, batcher_module, batcher_name, [result])

    async def push(self,
                   job_tag: str,
                   context: TransportContext,
                   batcher: Optional[BatcherConfig] = None,
                   adapter=None,
                   options: Optional[dict] = None,
                   on_error: Optional[Callable] = None,
                   queued: Optional[bool] = True
                   ) -> DispatchStatus:

        serialized_args = adapter.adapter_protocol.publish(self.args, on_error)
        self.args = adapter.adapter_protocol.consumer(serialized_args)

        serialized_kwargs = adapter.adapter_protocol.publish(self.kwargs, on_error)
        self.kwargs = adapter.adapter_protocol.consumer(serialized_kwargs)

        if on_error:
            on_error_function = lambda payload, e: on_error(payload, "No adapter", e)
        else:
            on_error_function = None

        assert isinstance(context, TransportContext)

        if not self._allow(context):
            return DispatchStatus(result=None, logs=log_handler.collection, status=403, error=None)

        try:

            logger.debug(f"Running inline. job tag: {job_tag}")
            result = await self._invoke(batcher, context)
            return DispatchStatus(result=result, logs=log_handler.collection, status=200, error=None)


        # On connection error
        except Exception as e:
            fallback.set_error_mode(str(e))
            logger.error(str(e))
            publish_payload = PublishPayload(
                capsule=self,
                job_tag=job_tag,
                context=context,
                headers={},
                options={}
            )
            if on_error_function:
                try:
                    on_error_function(publish_payload)
                except Exception as e:
                    logger.error(str(e))
            return DispatchStatus(result=None, logs=log_handler.collection, status=200, error=None)



from typing import Optional, Dict

from pydantic import BaseModel

RECORDS = "records"


class TransportContext(BaseModel):
    production: bool
    tenant: str
    trace_id: str
    properties: Optional[Dict] = {}

    def as_context(self) -> dict:
        return self.model_dump(mode='json', exclude={"properties": ...})

    @staticmethod
    def build(context, params=None):
        if params:
            params = {}
        return TransportContext(tenant=context.tenant,
                                production=context.production,
                                trace_id=context.trace_id,
                                properties=params)


import orjson

from durable_dot_dict.dotdict import DotDict
from pydantic import BaseModel


def fallback(obj):
    if isinstance(obj, DotDict):
        return obj.to_dict()
    if isinstance(obj, BaseModel):
        return obj.model_dump(mode="json")
    return str(obj)


class JsonSerializer:

    @staticmethod
    def serialize(obj):
        bytes = orjson.dumps(obj, default=fallback)
        return bytes.decode()

    @staticmethod
    def deserialize(message):
        return orjson.loads(message)
